/*
  D4

machine
  0 -- 1 -- 3
       |
       2
display
  1 -- 2 -- 4
       |
       3
 */

#include "qchar.h"

extern Wttable highestweight;

void show_vertex()
{
  fprintf(stderr,"   1 -- 2 -- 4\n"
                 "        |\n"
	  "        3\n");
}

Vertex no_vertex(Vertex no)
{

#ifdef DEBUG
  if (no > RANK || no < 1) {
      fprintf(stderr, "vertex must be between 1 and %d.\n",RANK);
      exit(1);
  }
#endif

  switch(no) {
  case 1: return(0);
	break;
  case 2: return(1);
	break;
  case 3: return(2);
	break;
  case 4: return(3);
	break;
  }  
}

Vertex iota(Vertex no)
{
  return(no);
}

short int exponent(Wttable vtable, Vertex vertex, Eigen eigen)
{
  short int e;

  e = highestweight[vertex][eigen];
  if (eigen < EIG_MAX - 1) e -= vtable[vertex][eigen+1];
  if (eigen > 0) e -= vtable[vertex][eigen-1];

  switch(vertex) {
  case 0: return(e + vtable[1][eigen]);
    break;
  case 1: return(e + vtable[0][eigen] + vtable[2][eigen] + vtable[3][eigen]);
    break;
  case 2: return(e + vtable[1][eigen]);
    break;
  case 3: return(e + vtable[1][eigen]);
    break;
  }
}  

#ifdef COMMENT
mul_A_sub(Wttable table, Vertex vertex, Eigen eigen,short int add)
     /*     Wttable table;
     short int vertex;
     short int eigen;
     short int add; */

{
  switch(vertex) {
  case 0: dainyuu(table,1,eigen+1,add);
    break;
  case 1: dainyuu(table,0,eigen+1,add);
    dainyuu(table,2,eigen+1,add);
    dainyuu(table,3,eigen+1,add);
    break;
  case 2: dainyuu(table,1,eigen+1,add);
    break;
  case 3: dainyuu(table,1,eigen+1,add);
    break;
  }
}  

int longest[12] = {1,3,2,1,3,2,4,2,1,3,2,4};
#endif

boolean Cinv(Weight weight1, Weight weight2)
     /* Calculate weight2 = C^{-1} weight1.
	Return FALSE if it is not integer valued. */
{
  Vertex vertex;
  short int s;

  weight2[0] = weight1[0]*2 + weight1[1]*2 + weight1[2] + weight1[3];
  if ((weight2[0] % 2 != 0) || weight2[0] < 0) return(FALSE);
  else weight2[0] = weight2[0]/2;

  weight2[1] = weight1[0] + weight1[1]*2 + weight1[2] + weight1[3];
  if (weight2[1] < 0) return(FALSE);

  weight2[2] = weight1[0] + weight1[1]*2 + weight1[2]*2 + weight1[3];
  if ((weight2[2] % 2 != 0) || weight2[2] < 0) return(FALSE);
  else weight2[2] = weight2[2]/2;

  weight2[3] = weight1[0] + weight1[1]*2 + weight1[2] + weight1[3]*2;
  if ((weight2[3] %2 != 0) || weight2[3] < 0) return(FALSE);
  else weight2[3] = weight2[3]/2;

  return(TRUE);

}

/*
> C := Matrix([[2,-1,0,0],[-1,2,-1,-1],[0,-1,2,0],[0,-1,0,2]]);
                               [ 2    -1     0     0]
                               [                    ]
                               [-1     2    -1    -1]
                          C := [                    ]
                               [ 0    -1     2     0]
                               [                    ]
                               [ 0    -1     0     2]

> inverse(C);
                           [ 1     1    1/2    1/2]
                           [                      ]
                           [ 1     2     1      1 ]
                           [                      ]
                           [1/2    1     1     1/2]
                           [                      ]
                           [1/2    1    1/2     1 ]
*/
