/*
  E6

  no of vertices.
  
machine:
  0 -- 2 -- 3 -- 4 -- 5
            |
    	    1
	    
display:
  1 -- 2 -- 3 -- 4 -- 5
            |
     	    6

 */

#include "qchar.h"

extern Wttable highestweight;

void show_vertex()
{
  fprintf(stderr,"   1 -- 2 -- 3 -- 4 -- 5\n"
                 "             |\n"
	  "             6\n");
}

Vertex no_vertex(Vertex no)
{

#ifdef DEBUG
  if (no > RANK || no < 1) {
      fprintf(stderr, "vertex must be between 1 and %d.\n",RANK);
      exit(1);
  }
#endif

  switch(no) {
  case 1: return(0);
	break;
  case 2: return(2);
	break;
  case 3: return(3);
	break;
  case 4: return(4);
	break;
  case 5: return(5);
	break;
  case 6: return(1);
	break;
  }  
}

Vertex iota(Vertex no)
     /* Dynkin diagram involution */
{

#ifdef DEBUG
  if (no >= RANK) {
      fprintf(stderr, "vertex must be between 1 and %d.\n",RANK);
      exit(1);
  }
#endif

  switch(no) {
  case 0: return(5);
    break;
  case 1: return(1);
    break;
  case 2: return(4);
	break;
  case 3: return(3);
	break;
  case 4: return(2);
	break;
  case 5: return(0);
	break;
  }  
}

/*
short int vertex_no(Vertex vertex)
{
  switch(vertex) {
  case 0: return(1);
	break;
  case 1: return(6);
	break;
  case 2: return(2);
	break;
  case 3: return(3);
	break;
  case 4: return(4);
	break;
  case 5: return(5);
	break;
  }
}
*/

short int exponent(Wttable vtable, Vertex vertex, Eigen eigen)
{
  short int e;

  e = highestweight[vertex][eigen];
  if (eigen < EIG_MAX - 1) e -= vtable[vertex][eigen+1];
  if (eigen > 0) e -= vtable[vertex][eigen-1];

  switch(vertex) {
  case 0: return(e + vtable[2][eigen]);
    break;
  case 1: return(e + vtable[3][eigen]);
    break;
  case 2: return(e + vtable[0][eigen] + vtable[3][eigen]);
    break;
  case 3: return(e + vtable[1][eigen] + vtable[2][eigen] + vtable[4][eigen]);
    break;
  case 4: return(e + vtable[3][eigen] + vtable[5][eigen]);
    break;
  case 5: return(e + vtable[4][eigen]);
    break;
  }
}  

#ifdef COMMENT
void mul_A_sub(Wttable table, Vertex vertex, Eigen eigen,short int add)
     /*     Wttable table;
     short int vertex;
     short int eigen;
     short int add; */

{
  switch(vertex) {
  case 0: dainyuu(table,2,eigen+1,add);
    break;
  case 1: dainyuu(table,3,eigen+1,add);
    break;
  case 2: dainyuu(table,0,eigen+1,add);
    dainyuu(table,3,eigen+1,add);
    break;
  case 3: dainyuu(table,1,eigen+1,add);
    dainyuu(table,2,eigen+1,add);
    dainyuu(table,4,eigen+1,add);
    break;
  case 4: dainyuu(table,3,eigen+1,add);
    dainyuu(table,5,eigen+1,add);
    break;
  case 5: dainyuu(table,4,eigen+1,add);
    break;
  }
}  

int longest[36] = {1, 2, 3, 1, 4, 2, 3, 1, 4, 3, 5, 4, 2, 3, 1, 4, 3, 5, 4, 2, 6, 5, 4, 2, 3, 1, 4, 3, 5, 4, 2, 6, 5, 4, 3, 1};
#endif

boolean Cinv(Weight weight1, Weight weight2)
     /* Calculate weight2 = C^{-1} weight1.
	Return FALSE if it is not integer valued. */
{
  Vertex vertex;
  short int s;

  weight2[0] = weight1[0]*4 + weight1[1]*3 + weight1[2]*5 + weight1[3]*6 + weight1[4]*4 + weight1[5]*2;
  if ((weight2[0] % 3 != 0) || weight2[0] < 0) return(FALSE);
  else weight2[0] = weight2[0]/3;

  weight2[1] = weight1[0] + weight1[1]*2 + weight1[2]*2 + weight1[3]*3 + weight1[4]*2 + weight1[5];
  if (weight2[1] < 0) return(FALSE);

  weight2[2] = weight1[0]*5 + weight1[1]*6 + weight1[2]*10 + weight1[3]*12 + weight1[4]*8 + weight1[5]*4;
  if ((weight2[2] % 3 != 0) || weight2[2] < 0) return(FALSE);
  else weight2[2] = weight2[2]/3;

  weight2[3] = weight1[0]*2 + weight1[1]*3 + weight1[2]*4 + weight1[3]*6 + weight1[4]*4 + weight1[5]*2;
  if (weight2[3] < 0) return(FALSE);

  weight2[4] = weight1[0]*4 + weight1[1]*6 + weight1[2]*8 + weight1[3]*12 + weight1[4]*10 + weight1[5]*5;
  if ((weight2[4] % 3 != 0) || weight2[4] < 0) return(FALSE);
  else weight2[4] = weight2[4]/3;

  weight2[5] = weight1[0]*2 + weight1[1]*3 + weight1[2]*4 + weight1[3]*6 + weight1[4]*5 + weight1[5]*4;
  if ((weight2[5] % 3 != 0) || weight2[5] < 0) return(FALSE);
  else weight2[5] = weight2[5]/3;

  return(TRUE);

}

/*

                         [ 2     0    -1     0     0     0]
                         [                                ]
                         [ 0     2     0    -1     0     0]
                         [                                ]
                         [-1     0     2    -1     0     0]
                    C := [                                ]
                         [ 0    -1    -1     2    -1     0]
                         [                                ]
                         [ 0     0     0    -1     2    -1]
                         [                                ]
                         [ 0     0     0     0    -1     2]

> inverse(C);
                    [4/3    1    5/3     2    4/3     2/3]
                    [                                    ]
                    [ 1     2     2      3     2       1 ]
                    [                                    ]
                    [5/3    2    10/3    4    8/3     4/3]
                    [                                    ]
                    [ 2     3     4      6     4       2 ]
                    [                                    ]
                    [4/3    2    8/3     4    10/3    5/3]
                    [                                    ]
                    [2/3    1    4/3     2    5/3     4/3]


*/
