/*
  E7

  no of vertices:
  
machine
  0 -- 2 -- 3 -- 4 -- 5 -- 6
            |
    	    1
display
  1 -- 2 -- 3 -- 4 -- 5 -- 6
            |
            7
 */

#include "qchar.h"

extern Wttable highestweight;

void show_vertex()
{
  fprintf(stderr,"   1 -- 2 -- 3 -- 4 -- 5 -- 6\n"
                 "             |\n"
	  "             7\n");
}

Vertex no_vertex(Vertex no)
{
  if (no > RANK || no < 1) {
      fprintf(stderr, "vertex must be between 1 and %d.\n",RANK);
      exit(1);
  }
  switch(no) {
  case 1: return(0);
	break;
  case 2: return(2);
	break;
  case 3: return(3);
	break;
  case 4: return(4);
	break;
  case 5: return(5);
	break;
  case 6: return(6);
	break;
  case 7: return(1);
	break;
  }  
}

Vertex iota(Vertex no)
{
  return(no);
}

/*
short int vertex_no(Vertex vertex)
{
  switch(vertex) {
  case 0: return(1);
	break;
  case 1: return(7);
	break;
  case 2: return(2);
	break;
  case 3: return(3);
	break;
  case 4: return(4);
	break;
  case 5: return(5);
	break;
  case 6: return(6);
	break;
  }
}
*/

short int exponent(Wttable vtable, Vertex vertex, Eigen eigen)
{
  short int e;

  e = highestweight[vertex][eigen];
  if (eigen < EIG_MAX - 1) e -= vtable[vertex][eigen+1];
  if (eigen > 0) e -= vtable[vertex][eigen-1];

  switch(vertex) {
  case 0: return(e + vtable[2][eigen]);
    break;
  case 1: return(e + vtable[3][eigen]);
    break;
  case 2: return(e + vtable[0][eigen] + vtable[3][eigen]);
    break;
  case 3: return(e + vtable[1][eigen] + vtable[2][eigen] + vtable[4][eigen]);
    break;
  case 4: return(e + vtable[3][eigen] + vtable[5][eigen]);
    break;
  case 5: return(e + vtable[4][eigen] + vtable[6][eigen]);
    break;
  case 6: return(e + vtable[5][eigen]);
    break;
  }
}

#ifdef COMMENT
void mul_A_sub(Wttable table, Vertex vertex,Eigen eigen,short int add)
     /* Wttable table;
     Vertex vertex;
     Eigen eigen;
     short int add; */

{
  switch(vertex) {
  case 0: dainyuu(table,2,eigen+1,add);
    break;
  case 1: dainyuu(table,3,eigen+1,add);
    break;
  case 2: dainyuu(table,0,eigen+1,add);
    dainyuu(table,3,eigen+1,add);
    break;
  case 3: dainyuu(table,1,eigen+1,add);
    dainyuu(table,2,eigen+1,add);
    dainyuu(table,4,eigen+1,add);
    break;
  case 4: dainyuu(table,3,eigen+1,add);
    dainyuu(table,5,eigen+1,add);
    break;
  case 5: dainyuu(table,4,eigen+1,add);
    dainyuu(table,6,eigen+1,add);
    break;
  case 6: dainyuu(table,5,eigen+1,add);
    break;
  }
}  
#endif

boolean Cinv(Weight weight1, Weight weight2)
     /* Calculate weight2 = C^{-1} weight1.
	Return FALSE if it is not integer valued. */
{
  Vertex vertex;
  short int s;

  weight2[0] = weight1[0]*2 + weight1[1]*2 + weight1[2]*3 + weight1[3]*4 + weight1[4]*3 + weight1[5]*2 + weight1[6];
  if (weight2[0] < 0) return(FALSE);

  weight2[1] = weight1[0]*4 + weight1[1]*7 + weight1[2]*8 + weight1[3]*12 + weight1[4]*9 + weight1[5]*6 + weight1[6]*3;
  if ((weight2[1] % 2) || (weight2[1] < 0)) return(FALSE);
  else weight2[1] = weight2[1] / 2;

  weight2[2] = weight1[0]*3 + weight1[1]*4 + weight1[2]*6 + weight1[3]*8 + weight1[4]*6 + weight1[5]*4 + weight1[6]*2;
  if (weight2[2] < 0) return(FALSE);

  weight2[3] = weight1[0]*4 + weight1[1]*6 + weight1[2]*8 + weight1[3]*12 + weight1[4]*9 + weight1[5]*6 + weight1[6]*3;
  if (weight2[3] < 0) return(FALSE);

  weight2[4] = weight1[0]*6 + weight1[1]*9 + weight1[2]*12 + weight1[3]*18 + weight1[4]*15 + weight1[5]*10 + weight1[6]*5;
  if ((weight2[4] % 2 != 0) || weight2[4] < 0) return(FALSE);
  else weight2[4] = weight2[4] / 2;

  weight2[5] = weight1[0]*2 + weight1[1]*3 + weight1[2]*4 + weight1[3]*6 + weight1[4]*5 + weight1[5]*4 + weight1[6]*2;
  if (weight2[5] < 0) return(FALSE);

  weight2[6] = weight1[0]*2 + weight1[1]*3 + weight1[2]*4 + weight1[3]*6 + weight1[4]*5 + weight1[5]*4 + weight1[6]*3;
  if ((weight2[6] % 2 != 0) || weight2[6] < 0) return(FALSE);
  else weight2[6] = weight2[6]/2;

  return(TRUE);

}

/*
                      [ 2     0    -1     0     0     0     0]
                      [                                      ]
                      [ 0     2     0    -1     0     0     0]
                      [                                      ]
                      [-1     0     2    -1     0     0     0]
                      [                                      ]
                 C := [ 0    -1    -1     2    -1     0     0]
                      [                                      ]
                      [ 0     0     0    -1     2    -1     0]
                      [                                      ]
                      [ 0     0     0     0    -1     2    -1]
                      [                                      ]
                      [ 0     0     0     0     0    -1     2]

> inverse(C);
                   [2     2     3     4     3      2     1 ]
                   [                                       ]
                   [2    7/2    4     6    9/2     3    3/2]
                   [                                       ]
                   [3     4     6     8     6      4     2 ]
                   [                                       ]
                   [4     6     8    12     9      6     3 ]
                   [                                       ]
                   [3    9/2    6     9    15/2    5    5/2]
                   [                                       ]
                   [2     3     4     6     5      4     2 ]
                   [                                       ]
                   [1    3/2    2     3    5/2     2    3/2]

*/
