#include "stdlib.h"
#include "ctype.h"
#include "qchar.h"

extern void memory_usage();
extern unsigned short int depth;
extern Poincare pone, pzero; /* constant Poincare polynomials */

/* subroutines handling Poincare polynomials */

void setp10()
/* Set pone = 1, pzero = 0 */
{
#ifndef EULER
  unsigned short int t;
#endif

#ifndef EULER
  for (t=0; t < COMP_DIM_MAX; t++)
    pone[t] = pzero[t] = 0;
  pone[0] = 1;
#else
  pone = 1; pzero = 0;
#endif
}

void clear_weight_P(Poincare weight[])
     /* clear the I-tuple of Poincare polynomials */
{
  Vertex i;
#ifdef EULER
  for (i=0; i<RANK; i++)
    weight[i] = 0;
#else
  unsigned short int t;

  for (i=0; i<RANK; i++)
    for (t=0; t < COMP_DIM_MAX; t++)
      weight[i][t] = 0;
#endif
}

#ifndef EULER
boolean equal_P(Poincare p1, Poincare p2)
{
  unsigned short int t = 0;

  for (t = 0; t < COMP_DIM_MAX; t++)
    if (p1[t] != p2[t]) return(FALSE);
  return(TRUE);
}
#endif

void display_P_tofile(FILE *file, Poincare p)
{
#ifdef EULER
  fprintf(file, "%d", p);
#else
  unsigned short int t = 0;

  fprintf(file, "%d", p[0]);
  
  while (++t < COMP_DIM_MAX) {
    if (p[t] == 1)
      fprintf(file, " + t^%d", 2*t);
    else if (p[t] != 0)
      fprintf(file, " + %d*t^%d", p[t], 2*t);
  }
#endif
}  

void display_PL_tofile(FILE *file, PoincareLong p)
{
#ifdef EULER
  fprintf(file, "%d", p);
#else
  unsigned short int t = 0;

  fprintf(file, "%d", p[0]);
  
  while (++t < COMP_DIM_MAX) {
    if (p[t] == 1)
      fprintf(file, " + t^%d", 2*t);
    else if (p[t] != 0)
      fprintf(file, " + %d*t^%d", p[t], 2*t);
  }
#endif
}  

#ifndef EULER
void shift_P(Poincare p, unsigned short int s)
{
  unsigned short int t = COMP_DIM_MAX;

  while (--t > COMP_DIM_MAX - s)
    if (p[t] != 0) {
      /*
      fprintf(stderr,\
	      "Out of range in Poincare polynomial at depth = %d. "
	      "Redefine COMP_DIM_MAX larger\n", depth);
      memory_usage();
	      */
      fprintf(stderr,\
	      "Out of range in Poincare polynomial "
	      "Redefine COMP_DIM_MAX = %d larger\n", COMP_DIM_MAX);
      exit(1);
    }

  if (p[t] != 0) {
    /*
    fprintf(stderr,\
	    "Out of range in Poincare polynomial at depth = %d. "
	    "Redefine COMP_DIM_MAX larger\n", depth);
    memory_usage();
	      */
    fprintf(stderr,\
            "Out of range in Poincare polynomial "
	    "Redefine COMP_DIM_MAX = %d larger\n", COMP_DIM_MAX);
    exit(1);
  }

  while (t-- > 0)
    p[t+s] = p[t];

  for (t=s; t > 0; t--)
    p[t-1] = 0;
}
#endif

#ifndef EULER
void add_P(Poincare p1, Poincare p2, Poincare p3)
{
  unsigned short int t;

  for (t=0; t<COMP_DIM_MAX; t++) {
    p3[t] = p1[t] + p2[t];
  }
}
#endif

#ifndef EULER
void subadd_P(Poincare p1, Poincare p2)
{
  unsigned short int t;

  for (t=0; t<COMP_DIM_MAX; t++) {
    p1[t] += p2[t];
  }
}

void subadd_PL(PoincareLong p1, PoincareLong p2)
{
  unsigned short int t;

  for (t=0; t<COMP_DIM_MAX; t++) {
    p1[t] += p2[t];
  }
}
#endif

#ifndef EULER
void subtr_P(Poincare p1, Poincare p2, Poincare p3)
{
  unsigned short int t;

  for (t=0; t<COMP_DIM_MAX; t++) {
    p3[t] = p1[t] - p2[t];
  }
}
#endif

#ifndef EULER
void mul_P(Poincare p1, Poincare p2, Poincare p3)
{
  unsigned short int s,t;
  Poincare p4;

  for (t=0; t<COMP_DIM_MAX;t++)
    p3[t] = 0;

  for (s=0; s<COMP_DIM_MAX;s++)
    if (p2[s] != 0) {
      for (t=0; t<COMP_DIM_MAX;t++)
	p4[t] = p1[t] * p2[s];
      shift_P(p4,s);
      add_P(p3,p4,p3);
    }
}
#endif

/* q-binomial coefficients */
#ifndef EULER
void binom(int n,int r,Poincare p)
{
  unsigned short int i;
  Poincare p1, p2;

  p[0] = 1;
  for (i=1; i<COMP_DIM_MAX; i++)
    p[i] = 0;

  if (n<0 || r<0 || n<r) return;
  if (n==r || r==0) return;
  
  binom(n-1,r-1,p1);
  binom(n-1,r,p2);
  shift_P(p2,r);
  add_P(p1,p2,p);
  return;
}
#else
/* usual binomial coefficients */
void binom(int n,int r, Poincare *p)
{
  Poincare p1,p2;
  *p = 1;
  
  if (n<0 || r<0 || n<r) return;
  if (n==r || r==0) return;
  binom(n-1,r-1,&p1);
  binom(n-1,r,&p2);
  *p = p1 + p2;
  return;
}
#endif

void byte2poincare(Qweight_cont data, Poincare p[])
{
  Vertex vertex;

  for (vertex=0; vertex<RANK; vertex++)
    if (bit_P(data, vertex) == 0)
      subst_P(p[vertex], pzero);
    else
      subst_P(p[vertex], pone);
}
