/* Computation of t-analog of q-character
    (original q-character was introduced by Frenkel-Reshetkhin)
                    programmed by Hiraku Nakajima
*/

#include "stdlib.h"
#include "ctype.h"
#include "sys/types.h"
#include "qchar.h"
#include "time.h"
#include <unistd.h>

extern boolean equal_weight(Wttable table, Weight weight);
extern unsigned int exp_max(Wttable table);
extern boolean is_i_dom(Wttable table, Vertex vertex);
extern boolean is_l_dominant(Wttable table);
extern boolean is_i_regdom(Wttable table,Vertex vertex);
extern int weight_i(Vertex vertex, Wttable table);
extern boolean is_dominant(Wttable table);
extern short int exponent(Wttable vtable, Vertex vertex, Eigen eigen);
extern void mul_A_sub(Wttable table,Vertex vertex,\
		      Eigen eigen,short int add);
extern void table2weight(Wttable table, Weight weight);
extern void clear_weight_P(Poincare weight[]);

extern void get_qwt(unsigned short int ii, unsigned int address, void (*p_func)(unsigned int));
/* p_func : function pointer 'develop_or_move' or 'move_dep_zero' or
   'just_display' */

#ifndef EULER
extern boolean equal_P(Poincare p1, Poincare p2);
/* extern void subst_P(Poincare p1, Poincare p2); */
extern void add_P(Poincare p1, Poincare p2, Poincare p3);
extern void subtr_P(Poincare p1, Poincare p2, Poincare p3);
extern void subadd_P(Poincare p1, Poincare p2);
extern void mul_P(Poincare p1, Poincare p2, Poincare p3);
#endif

#ifndef EULER
extern void binom(int n,int r,Poincare p);
#else
extern void binom(int n,int r,Poincare *p);
#endif

extern void display_P_tofile(FILE *file, Poincare p);
#define display_P(p) display_P_tofile(stdout, p)
extern void byte2poincare(Qweight_cont data, Poincare p[]);

extern void vtable2qwt(Wttable vtable, Qweight qwt);
extern void qwt2vtable(Qweight qwt, Wttable vtable);
extern void display_table_tofile(FILE *file, Wttable table);
#define display_table(table) display_table_tofile(stdout, table)
#define display_table_tostderr(table) display_table_tofile(stderr, table)
extern void display_table_tostr(Wttable table, char *string);
/* extern void sub_table(Wttable table1, Wttable table2); */
extern void sub_qwt(Qweight qwt1, Qweight qwt2);
extern unsigned short int qwt_insert(Qweight qwt, Eigen eigen,\
				     Vertex vertex);

#ifndef EULER
extern void fread_P(FILE *file, Poincare p);
extern void fwrite_P(FILE *file, Poincare p);
#endif

extern void display_1qwt_tostderr(Qweight_cont data);
extern void display_qwt_tostderr(Qweight qwt);
extern void u2v_sub(Eigen eigen, Wttable table);
extern void u2v(Wttable table);
extern void V(Wttable table, Wttable vtable);
extern void U(Wttable vtable, Wttable table);
extern unsigned int dim(Wttable table, Wttable vtable);

extern void read_hwt(char *str);
extern void str2weight(char *str);
extern void dim_range(char *str, struct Range *range);

/* Global variables */

/* program name, e.g., E6, E7 */
char *prog;

/* 
   Qold : array of l-weights to be developed
   Qnew : array of found l-weights

   They have a tree structure:

   data is a Qweight_cont, i.e., a vector (vertex, eigen)
                                 and corresponds to A_{vertex,eigen}^{-1}
				 (not to Y_{vertex,eigen}

   data -- data -- data --- data (= LAST)
    |       |       | (next=MINUS_ONE) 
    |       | next  X
   data     |
           data -- data --
            |       |
            |       | next
            |       |

   If next == MINUS_ONE, it means no next data.
   If data == LAST, it is a tail of the tree.
     In this case, the succeeding (data, next) have a special meaning.
     If data != LAST, it is an I-tuple of Poincare polynomials, which
       are either 0 or 1. (In this case `next' is NOT used.)
     If some Poincare polynomials are neither 0 nor 1,
        then data == LAST and next = a number of the monomial,
	 used for index for the array Mult_*[ ], which stores
	 Poincare polynomials.

*/

#ifdef COMMENT
#ifdef __GNUC__
struct Qweights {
  Qweight_cont data;
  unsigned int next __attribute__ ((packed));
};
#else
/* #pragma pack(2) */
/* #pragma pack makes bus error... */
struct Qweights {
  Qweight_cont data;
  unsigned int next;
};
/* #pragma pack() */
#endif
/* struct Qweights *Qnew; */
/* struct Qweights Qnew[LEN_MAX]; */
/* struct Qweights *Qnew_end; */
/* struct Qweights *Qold; */
/* struct Qweights Qold[LEN_MAX]; */
/* struct Qweights *Qold_end; */
#endif

Qweight_cont Qnew_data[LEN_MAX];
unsigned int Qnew_end;
unsigned int Qnew_next[LEN_MAX];

Qweight_cont Qold_data[LEN_MAX];
unsigned int Qold_end;
unsigned int Qold_next[LEN_MAX];

Qweight qwt_develop, qwt_tmp;
/* Qweight_cont *qwt_develop; */
/* Qweight_cont *qwt_move; */

struct Qwt_find {
  Qweight_cont *qwt;
  unsigned int address;
};

Poincare pone, pzero; /* constant Poincare polynomials */

unsigned int find_term_no = 0;
unsigned long long int count = 0;

unsigned short int depth = 0;

Poincare Mult_new[MAX_NUM_TERMS][RANK];
Poincare Mult_old[MAX_NUM_TERMS][RANK];

FILE *outfile, *outfile2;

#ifdef E6
extern Vertex iota(Vertex no);
#else
#define iota(vertex) (vertex)
#endif

/* variables handling options */

Weight WEIGHT;
Wttable highestweight;
Qweight hwt;
struct Range DIM_range;
int opt_max_exp;
boolean IS_L_DOM, IS_DOM, EQ_WEIGHT;
boolean print_V = FALSE, print_depth = FALSE, write_file = FALSE,
  read_file = FALSE, write_file2 = FALSE, read_file2 = FALSE;

/* variables to show memory usage */
unsigned int len_max = 0;

#ifdef DEBUG
unsigned int max_exp = 0;
unsigned int MAX_exp = 0;
unsigned int max_dim = 0;
unsigned int MAX_dim = 0;
unsigned int max_num_terms = 0;
/* unsigned int no_terms; */
#ifdef EULER
unsigned int max_euler = 0;
unsigned int MAX_euler = 0;
#endif
unsigned int non_one = 0;
unsigned int max_non_one = 0;
#endif

/* Subroutines for option handlings */
void usage(void)
{
  fprintf(stderr,"Usage:  %s [-options] l-highest_weight\n"
	  "  l-highest weight is a sequence of a triple = vertex,eigen,expo\n"
	  "  e.g. 1,0,1,1,1,2 = Y_{1,a} Y_{1,qa}^2 \n"
	  "Options:\n"
	  "   -D#[,#] : print if dim of graded quiver variety\n"
	  "             given by a l-weight is in #<= dim <= #."
	  " (Print also dimension.)\n"
	  "   -V      : print V of l-weights\n"
	  "   -d      : print dimension of V (= depth)\n"
	  "   -m#     : print l-weights with max of exp >= #\n"
	  "   -p      : print dominants l-weights only\n"
	  "   -w#,#,#,...,# (rank times)\n"
	  "           : print l-weights with given weights "
	  "(# can be a negative integer)\n"
	  "   -f      : write data to files.\n" 
	  "   -F[#]   : read data from files.\n"
	  "             If # is set, print l-weights with depth=#.\n"
	  "   -b[#]   : write temporary data to files.\n"
	  "             If # is set, stop at stop=#. Otherwise, continue to the end.\n"
	  "             Use together with the option -B to stop the program on the way.\n"
	  "   -B#     : read the file of temporary data with depth = #,\n"
	  "             and start from there.\n"
	  "   -h      : this help\n"
	  "REMARK: become super user and unlimit stacksize, for some cases\n"
	  "Vertex Numbering:\n",
	  prog);
  show_vertex();
  exit(1);
}

/* Memory Use */
void memory_usage()
{
  fprintf(stderr,
	  "*** Memory Use in one step = fixed depth + developed terms ***\n"
	  "         (Total # of l-weights = %llu)\n"
	  "  Sum of length of l-weights = %d\n"
#ifdef DEBUG
	  "  # of terms with non one coeff. = %d\n"
	  "  Max of exponents = %d\n"
#ifdef EULER
          "  Max of Euler Number = %d\n"
#endif
#endif
	  , count, Qnew_end
#ifdef DEBUG
	  , find_term_no, max_exp
#ifdef EULER
	  , max_euler
#endif
#endif
	  );
#ifdef DEBUG
  if (DIM_range.low != -1 || DIM_range.high != -1)
    fprintf(stderr,"  Max of dim of components = %d\n", max_dim);
#endif
}

boolean cond(Wttable table, unsigned int dim)
{

  if (!EQ_WEIGHT || equal_weight(table,WEIGHT)) {
    if (opt_max_exp == -1 || exp_max(table) >= (unsigned )opt_max_exp) {
      if (!IS_DOM || is_dominant(table)) {
	if (((int )dim >= DIM_range.low ||
	     DIM_range.low == -1) &&
	    (DIM_range.high == -1 || (int )dim <= DIM_range.high))
	  return(TRUE);
      }
    }
  }
  return(FALSE);
}

void add_newqwt(Qweight_cont *qwt)
{
  Qnew_data[Qnew_end] = *qwt;
  Qnew_next[Qnew_end++] = MINUS_ONE;

  if (*qwt != LAST) {
    add_newqwt(qwt+1);
  } else {
    Qnew_data[Qnew_end] = LAST; /* specify later !*/
    Qnew_next[Qnew_end++] = find_term_no;
  }
}

struct Qwt_find find(Qweight_cont *qwt, unsigned int address)
/* Find `qwt' in Qnew_data[address + ****].
   Return :
   find.qwt is such that *(find.qwt) was not found, but
                         *(find.qwt-1) was found.
			 (or `find.qwt' = qwt means nothing found.)
   find.address is such that
        Qnew_data[find.address-1] = *(find.qwt-1), or
	find.address = address if nothing found. */
{
  struct Qwt_find tmp;

#ifdef DEBUG
  if (*qwt == LAST) {
    fprintf(stderr,"Error in subroutine `find'.\n"
	      "qwt - qwt_tmp = %d, address = %d\n", \
	    qwt-qwt_tmp, address);
    exit(3);
  }
#endif

  while (*qwt != Qnew_data[address]) {
    if (Qnew_next[address] == MINUS_ONE) { /* when no more data */
      tmp.qwt = qwt;
      tmp.address = address;
      return(tmp);
    }
    address += Qnew_next[address];
  }
  /* find *qwt ! */
  if (*(qwt+1) != LAST) /* when *(qwt+1) is not LAST */
    return(find(qwt+1, address+1));

  tmp.qwt = qwt+1;
  tmp.address = address+1;

  return(tmp);
}

unsigned int find_add_newqwt(Qweight_cont *qwt, unsigned int address)
/* If qwt is already founded before, return its address.
   Otherwise add that data to Qnew and return MINUS_ONE. */
{

  while (*qwt != Qnew_data[address]) {
    if (Qnew_next[address] == MINUS_ONE) { /* when no more data */
      Qnew_next[address] = Qnew_end - address;

      add_newqwt(qwt);

      if (Qnew_end > LEN_MAX - 100) {
	fprintf(stderr,\
		"Out of Memory at depth = %d. Redefine LEN_MAX larger.\n",\
		depth);
	memory_usage();
	exit(2);
      }
      return(MINUS_ONE);
    }
    address += Qnew_next[address];
  }
  /* find *qwt ! */
  if (*qwt == LAST) { /* when qwt is the last */
    /* the monomial was already found before */
    return(address+1);
  } else {
    return(find_add_newqwt(qwt+1, address+1));
  }
}

void develop_sub(Wttable vtable, Vertex vertex, Qweight qwt, Eigen eigen,\
		 boolean ischanged, Poincare coeff, unsigned short int dep,\
		 unsigned short int start_qwt, unsigned int start_address)
{
  Qweight qwt2;
  unsigned short int i, j, l;
  unsigned int find_addr, find_no;
  short int ex;
  Vertex v;
  Poincare coeff2, coeff3;
  struct Qwt_find temp;
  Qweight_cont d;

  /* short cut */
  while (eigen < EIG_MAX && (ex = exponent(vtable, vertex, eigen)) == 0)
    eigen++;

  if (eigen == EIG_MAX) {
    /* we shall check whether a monomial is already found or not */
    if (ischanged) {
      /* we need to check only if the monomial is not the developed one. */

/*       vtable2qwt(vtable2,qwt_tmp); */
      
      sub_qwt(qwt_tmp, qwt);
#ifdef DEBUG2
      temp = find(qwt_tmp, 0);
      fprintf(stderr, "%d  %d  %d \n",\
	      temp.qwt - qwt_tmp, temp.address, Qnew_end);
      if (temp.qwt != qwt_tmp) {
	display_1qwt_tostderr(*(temp.qwt-1));
	fprintf(stderr, "  ");
	display_1qwt_tostderr(*temp.qwt);
	fprintf(stderr, "\n");
	display_1qwt_tostderr(Qnew_data[temp.address-1]);
	fprintf(stderr, "  ");
	display_1qwt_tostderr(Qnew_data[temp.address]);
	fprintf(stderr, "\n");
      }
#endif

/*       find_addr = find_add_newqwt(qwt_tmp, 0); */
      find_addr = find_add_newqwt(qwt_tmp+start_qwt, start_address);

      if (find_addr != MINUS_ONE) {
	/* If already found, 
	   then Mult_new[find_no][vertex] += coeff; */
	if (Qnew_data[find_addr] != LAST) {
	  if (equal_P(coeff, pone)
	      && bit_P(Qnew_data[find_addr], vertex) == 0) {
	    Qnew_data[find_addr] |= byte_P(vertex);
	  } else {
	    for (v=0; v<RANK; v++) {
	      if (bit_P(Qnew_data[find_addr], v) == 1)
		subst_P(Mult_new[find_term_no][v], pone);
	      else
		subst_P(Mult_new[find_term_no][v], pzero);
	    }
	    subadd_P(Mult_new[find_term_no][vertex], coeff);
	    Qnew_data[find_addr] = LAST;
	    Qnew_next[find_addr] = find_term_no++;
	  }
	} else {
	  find_no = Qnew_next[find_addr];
	  subadd_P(Mult_new[find_no][vertex], coeff);
	}

      } else {
	/* not found before */
	if (!equal_P(coeff, pone)) {
	  clear_weight_P(Mult_new[find_term_no]);
	  subst_P(Mult_new[find_term_no][vertex], coeff);
	  find_term_no++;
	} else
 	  Qnew_data[Qnew_end-1] = byte_P(vertex);
	
	/* Memory check */
	len_max = max(Qnew_end, len_max);
	if (len_max > LEN_MAX - 100) {
	  fprintf(stderr,\
		  "Out of Memory at depth = %d. "
		  "Redefine LEN_MAX larger.\n", depth); 
	  memory_usage();
	  exit(2);
	}
	
 	if ((unsigned int)MAX_NUM_TERMS + 10 < find_term_no) {
 	  fprintf(stderr,\
		  "Out of Memory at depth = %d. "
		  "Redefine MAX_NUM_TERMS larger.\n", depth);
 	  memory_usage();
 	  exit(2);
 	}
      }
    }
  } else { /* i.e., if (eigen != EIG_MAX) */
#ifdef DEBUG
/* probably this is not necessary */
    if (ex < 0) {
      fprintf(stderr,"Error in program: negative exponent in develop.\n"
	      "vertex = %d, eigen = %d\n",vertex+1, eigen);
      display_table_tostderr(vtable); fprintf(stderr, "\n");
      exit(3);
    }
#endif
    i = 0;
    sub_qwt(qwt2, qwt);
    /* sub_table(vtable3, vtable2); */

    l = qwt_len(qwt2);
    j = qwt_insert(qwt2, eigen+1, vertex);

    if (j != 0) {
      d = qwt2[j];
      qwt2[j] = LAST;
      temp = find(qwt2, 0);
      qwt2[j] = d;
    } else {
      temp.qwt = qwt2; temp.address = 0;
    }

#ifdef DEBUG
  if (l + (unsigned short int)ex >= DEPTH_MAX) {
    fprintf(stderr,\
	    "Out of range in a q-weight table at depth = %d. "
	    "length of qwt2 = %d. "
	    "exponent = %d.\n"
	    "Redefine DEPTH_MAX larger\n", depth, l, ex);
    display_qwt_tostderr(qwt2);
    memory_usage();
    exit(1);
  }
#endif

    do {
#ifndef EULER
      binom(ex,i,coeff2);
#else
      binom(ex,i,&coeff2);
#endif
      mul_P(coeff,coeff2,coeff3);
      develop_sub(vtable, vertex, qwt2, eigen+1, ischanged, coeff3, dep,\
		  temp.qwt-qwt2, temp.address);

/*       vtable3[vertex][eigen+1]++; */

      memmove(qwt2+j+i+1, qwt2+j+i, (l-j+1)*sizeof (Qweight_cont));
      qwt2[j+i] = ieigen(eigen+1) + vertex;

      ischanged = TRUE;
      dep++;
    } while(i++ < (unsigned short int)ex);
  }
}

#ifdef DEBUG
void develop_error(Vertex vertex, Wttable table, Poincare coeff,\
		   Poincare p[RANK])
{
  Vertex v;

  fprintf(stderr, "Error in algorithm at vertex %d. Coeff = ",
	  vertex+1);
  display_P_tofile(stderr, coeff);
  fprintf(stderr, "\n  Qweight = "); display_table_tostderr(table);
  fprintf(stderr, "\n  Virtual Poincare polynomials are:\n  ");
  for (v=0; v < RANK; v++) {
    display_P_tofile(stderr, p[v]);
    fprintf(stderr, "   ");
  }
  fprintf(stderr, "\n");
  exit(3);
}
#endif

void display(Wttable table, Wttable vtable, Poincare coeff)
/* display found l-weights ! */
{
  unsigned int d;

  if (DIM_range.low == -1 && DIM_range.high == -1) {
    d = 0;
  } else {
    d = dim(table,vtable);
#ifdef DEBUG
    max_dim = max(d,max_dim);
#endif
  }

#ifdef DEBUG
  max_exp = max(max_exp,exp_max(table));
#ifdef EULER
  max_euler = max(max_euler, coeff);
#endif
  if (!equal_P(coeff, pone)) non_one++;
#endif

  if (cond(table,d)) {
    printf("%llu, ",count);
    if (DIM_range.low != -1 || DIM_range.high != -1)
      printf("dim. = %d, ",d);

    printf("multi. = ");
    display_P(coeff);
    printf(", ");

    printf("qweight =");
    display_table(table);
    
    if (print_V) {
      printf(", V =");
      display_table(vtable);
    }

    printf("\n");
  }
}


void develop(unsigned int address)
{
  Poincare coeff, coeff2, p[RANK];
  Wttable table, vtable;  /* Wttable vtable2; */
  Vertex vertex;
  boolean already_i_dom = FALSE;
  unsigned int d;
  unsigned int _term_no;

  qwt2vtable(qwt_develop, vtable);
  U(vtable, table);

  if (depth == 0) subst_P(coeff, pone);

  if (Qold_data[address] == LAST) {
    _term_no = Qold_next[address];
    subst_IP(p, Mult_old[_term_no]);
  } else byte2poincare(Qold_data[address], p);

  for (vertex = 0; vertex < RANK; vertex++) {
    if (!is_i_dom(table,vertex)) {
      if (!already_i_dom) {
	subst_P(coeff, p[vertex]);
	already_i_dom = TRUE;
      }
#ifdef DEBUG
      else if (!equal_P(coeff, p[vertex])) 
	  develop_error(vertex, table, coeff, p);
#endif
    }
  }
  count++;

/* display found l-weights ! */

  display(table,vtable,coeff);

/* Next we develop i-dominant l-weights */

  for(vertex=0; vertex<RANK; vertex++)
    if ( is_i_regdom(table,vertex) && !equal_P(p[vertex],coeff) ) {
      subtr_P(coeff,p[vertex],coeff2);
/*       sub_table(vtable2,vtable); */
      develop_sub(vtable, vertex, qwt_develop, 0, FALSE, coeff2, 0,
		  0,0);
    }
}

unsigned short int depcal(Qweight qwt)
{
   unsigned short int i = 0;
   
   while (qwt[i] != LAST) 
     i++;
   return(i - depth);

}

void develop_or_move(unsigned int address)
{
  unsigned int _term_no;
  Vertex vertex;
  unsigned int find_addr, find_no;

  if (depcal(qwt_develop) == 0) {
    /* A monomial with depth = 0 is developed */
    develop(address);
  } else {
    /* A monomial with depth > 1 is moved to Qnew */
    find_addr = find_add_newqwt(qwt_develop, 0);

    if (find_addr == MINUS_ONE) { /* not found before */
      
      /* TEMPORARY */
      if (Qold_data[address] != LAST) {
	Qnew_data[Qnew_end-1] = Qold_data[address];
	Qnew_next[Qnew_end-1] = MINUS_ONE; /* delete find_term_no */
      } else {
	_term_no = Qold_next[address];
	subst_IP(Mult_new[find_term_no], Mult_old[_term_no]);

	find_term_no++;
      }
      
    } else {/* found before */
      /* The followings are simply
	 `Add I-tuple of Poincare polynomials for Qold
	   to I-tuple of Poincare polynomials for Qnew'.
	 The complexity comes from the case study
	  whether Poincare polynomials are in { 0,1 }, or not */

      if (Qnew_data[find_addr] == LAST) {
	find_no = Qnew_next[find_addr];

	if (Qold_data[address] == LAST) {
	  /* i.e., Qnew_data[find_addr] = LAST & Qold_data[address] = LAST */
	  _term_no = Qold_next[address];
	  for (vertex = 0; vertex < RANK; vertex++)
	    subadd_P(Mult_new[find_no][vertex], Mult_old[_term_no][vertex]);
	} else {
	  /* i.e., Qnew_data[find_addr] = LAST & Qold_data[address] != LAST */
	  for (vertex=0; vertex<RANK; vertex++) {
	    if (bit_P(Qold_data[address], vertex) == 1)
	      subadd_P(Mult_new[find_no][vertex], pone);
	  }

	}
      } else /* i.e., Qnew_data[find_addr] != LAST */
	if (Qold_data[address] != LAST)
	  if ((Qnew_data[find_addr] & Qold_data[address]) == 0)
	    Qnew_data[find_addr] |= Qold_data[address];
	  else { /* i.e., Qold_data[address] != LAST &
		            Qnew_data[find_addr] != LAST &
			    their bit intersection is nonzero */
	    for (vertex=0; vertex<RANK; vertex++) {
	      if (bit_P(Qnew_data[find_addr], vertex) == 1)
		subst_P(Mult_new[find_term_no][vertex], pone);
	      else
		subst_P(Mult_new[find_term_no][vertex], pzero);
	      if (bit_P(Qold_data[address], vertex) == 1)
		subadd_P(Mult_new[find_term_no][vertex], pone);
/* 	      else */
/* 		subadd_P(Mult_new[find_term_no][vertex], pzero); */
	    }
	    Qnew_data[find_addr] = LAST;
	    Qnew_next[find_addr] = find_term_no;
	    find_term_no++;
	  }
	else { /* i.e., Qold_data[address] = LAST &
		          Qnew_data[find_addr] != LAST */
	  _term_no = Qold_next[address];

	  for (vertex=0; vertex < RANK; vertex++) {
	    if (bit_P(Qnew_data[find_addr], vertex) == 1)
	      subst_P(Mult_new[find_term_no][vertex], pone);
	    else
	      subst_P(Mult_new[find_term_no][vertex], pzero);
	    subadd_P(Mult_new[find_term_no][vertex],\
		     Mult_old[_term_no][vertex]);
	  }
	  Qnew_data[find_addr] = LAST;
	  Qnew_next[find_addr] = find_term_no;
	  find_term_no++;
	}
    }
  }
}

/* The following subroutine is used to pick up l-weights with depth = 0. 
   It is almost the same as `develop_or_move' */

void move_dep_zero(unsigned int address)
{
  unsigned int _term_no;
  Vertex vertex;
  unsigned int find_addr, find_no;
  boolean already_i_dom = FALSE;
  Wttable table,vtable;

  if (depcal(qwt_develop) == 0) {
    /* A monomial with depth = 0 is written */
    find_addr = find_add_newqwt(qwt_develop, 0);

#ifdef DEBUG
    if (find_addr != MINUS_ONE) { /* was found before ! */
      fprintf(stderr, "Error ! Duplicate data was found.");
      fprintf(stderr, "\n  Qweight = "); display_qwt_tostderr(qwt_develop);
    }
#endif
      
    if (Qold_data[address] != LAST) {
      Qnew_data[Qnew_end-1] = Qold_data[address];
      Qnew_next[Qnew_end-1] = MINUS_ONE; /* delete find_term_no */
    } else {
      _term_no = Qold_next[address];

      /* We do not need to write Mult_new for all vertex. We write the
	 result to Mult_new[find_term_no][0]. */

      qwt2vtable(qwt_develop,vtable);
      U(vtable,table);

      for (vertex = 0; vertex < RANK; vertex++) {
	if (!is_i_dom(table,vertex)) {
	  if (!already_i_dom) {
	    subst_P(Mult_new[find_term_no][0], Mult_old[_term_no][vertex]);
	    already_i_dom = TRUE;
	  }
#ifdef DEBUG
	  else if (!equal_P(Mult_old[_term_no][vertex], Mult_new[find_term_no][0]))  
	  develop_error(vertex, table, Mult_old[_term_no][vertex], Mult_new[find_term_no]);
#endif
	}
      }

#ifdef DEBUG2
      if (equal_P(Mult_new[find_term_no][0],pzero))
	fprintf(stderr,"something strange !\n");
#endif
      
      find_term_no++;
    }
  }
}

void just_display(unsigned int address)
{
  Poincare coeff,p[RANK];
  Wttable table, vtable;
  Vertex vertex;
  boolean already_i_dom = FALSE;
  unsigned int _term_no;

  qwt2vtable(qwt_develop,vtable);
  U(vtable,table);

  if (depth == 0) subst_P(coeff,pone);

  if (Qold_data[address] == LAST) {
    _term_no = Qold_next[address];
    subst_P(coeff, Mult_old[_term_no][0]);
  } else {

    /* In this case, coeff must be pone */
    subst_P(coeff,pone);

    /* But we check it */
#ifdef DEBUG
    byte2poincare(Qold_data[address],p);
    for (vertex =0; vertex < RANK; vertex++) {
      if (!is_i_dom(table,vertex)) {
	if (!equal_P(coeff,p[vertex]))
	  develop_error(vertex,table,coeff,p);
      }
    }
#endif	
  }

  count++;

  display(table,vtable,coeff);
}

#ifdef COMMENT
void get_qwt(unsigned short int ii, unsigned int address, void (*p_func)(unsigned int))
     /* p_func : function pointer 'develop_or_move' or 'move_dep_zero' or
      'just_display' */
{

  do {
    
    if ((qwt_develop[ii] = Qold_data[address]) == LAST) {
      /* get qwt ! */ 
      /* develop_or_move(address+1); */
      (*p_func)(address+1);
      /* address is necessary only to get term_no. */
    } else {
      get_qwt(ii+1, address+1, p_func);
    }

    if (Qold_next[address] == MINUS_ONE) /* when there are no more data */
      break;

    address += Qold_next[address];
  } while(TRUE);
}
#endif

void debug_display(unsigned int address, unsigned int c)
{
  unsigned int i;
  Vertex vertex;

  fprintf(stderr,"\n***** depth %d *****************\n", depth);
  fprintf(stderr, "         next     data = l-weight     Poincare's\n");
  for (i=0; i < c; i++) {
    fprintf(stderr, "%4x = ", i);
    fprintf(stderr, "%8x   %4x    ", Qold_next[address], Qold_data[address]);
    display_1qwt_tostderr(Qold_data[address]);

    fprintf(stderr, "    ");
    if (Qold_data[address] == LAST)
      for(vertex=0;vertex < RANK; vertex++) {
	fprintf(stderr, "   ");
	display_P_tofile(stderr, Mult_old[Qold_next[address+1]][vertex]);
      }

    fprintf(stderr,"\n");
    address++;
  }
}

main(int argc, char *argv[])
{

  unsigned short int dep_s = DEPTH_MAX+1;

  unsigned int i,j,d;
/*   int a,c,i,j,d; */
  /* struct Qweight_find f; */
  Vertex vertex;
/*   unsigned int term_no; */
  unsigned int Qlength;
  Qweight_cont qtmp;

/*   Poincare coeff, coeff2; */
/*   Poincare *temp1, *temp2; */
/*   unsigned int c1,c2; */
/*   Qweight qwt; */
/*   Wttable table,table2; */

#ifndef EULER
  unsigned short int t;
#endif

  char *fileName, *fileName_base;
  char depthName[3];
  unsigned int develop_term_no = 0;

  Weight tmp_weight, tmp_weight2;

#ifdef DEBUG
  time_t jikan;
#endif

  fileName = (char *)malloc(40);
  fileName_base = (char *)malloc(40);

/*   Qnew = (struct Qweights *)malloc(LEN_MAX * sizeof (struct Qweights)); */
/*   Qold = (struct Qweights *)malloc(LEN_MAX * sizeof (struct Qweights)); */

/*   printf("size of (struct Qweights) = %d\n", sizeof (struct Qweights)); */
/*   printf("size of (Poincare) = %d\n", sizeof (Poincare)); */
/*   printf("size of (short int) = %d\n", sizeof (short int)); */
/*     printf("%d\n", 0xA000000); */
/*   exit(0); */

  /* Set pone = 1, pzero = 0 */
  setp10();

  DIM_range.low = -1;
  DIM_range.high = -1;
  opt_max_exp = -1;
  IS_L_DOM = FALSE;
  EQ_WEIGHT = FALSE;

  /* Option handling */

  prog = *argv++;
  while (--argc > 1) {
    if (argv[0][0] == '-') {
      switch (argv[0][1]) {
      case 'V':
	print_V = TRUE;
	break;
      case 'D':
	dim_range(*argv+2,&DIM_range);
	/* printf("%d - %d\n", DIM_range.low,DIM_range.high); */
	break;
      case 'd':
	print_depth = TRUE;
	break;
      case 'm':
	opt_max_exp = max(atoi(*argv+2),opt_max_exp);
	break;
      case 'p':
	IS_DOM = TRUE;
	break;
      case 'w':
	EQ_WEIGHT = TRUE;
	str2weight(*argv+2);
	/* for (i=0; i<RANK; i++)
	   printf("%d ",WEIGHT[i]);
	   printf("\n"); */
	break;
      case 'f':
	write_file = TRUE;
	break;
      case 'F':
	read_file = TRUE;
	if (*(*argv+2) != '\0')
	  if (isdigit(*(*argv+2))) {
	    depth = atoi(*argv+2);
	    dep_s = depth+1;
	    if (depth == 0) usage();
	  } else usage();
	break;
      case 'b':
	write_file2 = TRUE;
	if (*(*argv+2) != '\0')
	  if (isdigit(*(*argv+2))) {
	    dep_s = atoi(*argv+2);
	    if (dep_s == 0) usage();
	  } else usage();
	break;
      case 'B':
	read_file2 = TRUE;
	depth = atoi(*argv+2);
	if (depth == 0) usage();
	break;
      case 'h':
      default:
	usage();
      }
      argv++;
    }
    else usage();
  }

  /* we must have stop depth > start depth  */
  if (write_file2 && read_file2 && (depth >= dep_s)) {
      fprintf(stderr, "stop depth must be larger than start depth.\n");
      exit(1);
  }


  /* We cannot use (-B,-b, or -f) and -F simultaneously */
  if (read_file && (write_file || write_file2 || read_file2)) {
    fprintf(stderr, "cannot use (-B, -b or -f) and -F simultaneously.\n");
    exit(1);
  }

  /* read highest weight */
  if (*argv == NULL) usage();
  read_hwt(*argv);

  if (read_file && depth == 0) {
  /* Calculate dep_s = the depth when we end */
  qwt2weight(hwt,tmp_weight);
  for (vertex=0; vertex < RANK; vertex++) 
    tmp_weight2[vertex] = -tmp_weight[iota(vertex)];
  for (vertex=0; vertex < RANK; vertex++) 
    tmp_weight[vertex] -= tmp_weight2[vertex];
  Cinv(tmp_weight,tmp_weight2);
  dep_s = 0;
  for (vertex=0; vertex < RANK; vertex++)
    dep_s += tmp_weight2[vertex];

  dep_s++;
  }

  /* display highest weight */
  /*    to stderr */
  fprintf(stderr,"*** highest weight =");
  display_table_tostderr(highestweight);
  fprintf(stderr," ***\n");

  Qnew_data[0] = Qold_data[0] = LAST;
  Qnew_next[0] = Qold_next[0] = MINUS_ONE;

  Qnew_end = Qold_end = 0;

/*   add_newqwt(hwt, &Qold_end, 0); */
/*   Depth_old[0] = 0; */

  /* file name */
  strcpy(fileName_base, prog);
  display_table_tostr(highestweight,depthName);
  strcat(fileName_base, depthName);
  strcat(fileName_base, " ");

  /* read data from the temporary file */
  if (read_file2) { 
    strcpy(fileName, fileName_base);
    sprintf(depthName, "%d", depth);
    strcat(fileName, depthName);
    strcat(fileName, "_tmp");
    if ((outfile2 = fopen(fileName, "r")) == NULL) {
      fprintf(stderr, "cannot open file %s !\n", fileName);
      exit(2);
    }

    fread(&Qold_end, sizeof Qold_end, 1, outfile2);

#ifdef DEBUG2
    fprintf(stderr, "Qold_end = %d\n", Qold_end);
#endif

    for (i=0; i < Qold_end; i++) {
      fread(&qtmp, sizeof qtmp, 1, outfile2);
      d = qtmp >> 12;
      
#ifdef DEBUG2
      fprintf(stderr, "read %x", qtmp);
      fprintf(stderr, " (qtmp >> 12 = %x, ", d);
#endif

      Qold_data[i] = qtmp & LAST;
      if (d == 0) Qold_next[i] = MINUS_ONE;
      else Qold_next[i] = 0;
      for (j=0; j < d; j++) {
	fread(&qtmp, sizeof qtmp, 1, outfile2);
	Qold_next[i] += (qtmp << 8*j);
      }

#ifdef DEBUG2
      display_1qwt_tostderr(Qold_data[i]);
      fprintf(stderr, "), Qold_next[i] = %x\n", Qold_next[i]);
#endif

      if (Qold_data[i] == LAST) {
	i++;
	fread(&(Qold_data[i]), sizeof qtmp, 1, outfile2);
	if (Qold_data[i] == LAST)
	  fread(&(Qold_next[i]), sizeof (unsigned int), 1, outfile2);

#ifdef DEBUG2
	fprintf(stderr, " %x\n", Qold_data[i]);
	fprintf(stderr, " %x\n", Qold_next[i]);
#endif

      }
    }

    fread(&develop_term_no, sizeof develop_term_no, 1, outfile2);
    for (i=0; i < develop_term_no; i++) {
      for (vertex=0; vertex<RANK; vertex++) {
	fread_P(outfile2, Mult_old[i][vertex]);
      }
    }
      
    len_max = Qold_end;

    fclose(outfile2);

#ifdef DEBUG
    time(&jikan);
    fprintf(stderr, "We finish the file reading at %s", ctime(&jikan));

    max_num_terms = develop_term_no;
#endif

  }

  /* Main Routine */

  do {

    if (print_depth) 
      printf("\n***** depth %d *****************\n", depth);

#ifdef DEBUG2
    debug_display(0, Qold_end);
#endif

    if ((write_file2 || read_file) && dep_s == depth) break;

    if (write_file && depth != 0) {
      /* write data with depth = 0 to Qnew */
      get_qwt(0,0,move_dep_zero);

      /* file open */
      strcpy(fileName, fileName_base);
      sprintf(depthName, "%d", depth);
      strcat(fileName, depthName);

      if ((outfile = fopen(fileName, "w")) == NULL) {
	fprintf(stderr, "File open error !\n");
	exit(2);
      }
      
      fwrite(&Qnew_end, sizeof Qnew_end, 1, outfile);
      
      for (i=0; i < Qnew_end; i++) {
	if (Qnew_next[i] == MINUS_ONE) 
	  d = 0;
	else if ((Qnew_next[i] & 0xFF000000) != 0)
	  d = 4;
	else if ((Qnew_next[i] & 0x00FF0000) != 0)
	  d = 3;
	else if ((Qnew_next[i] & 0x0000FF00) != 0)
	  d = 2;
	else
	  d = 1;

	qtmp = ( (Qnew_data[i] & LAST) | (d << 12) );
	fwrite(&qtmp, sizeof qtmp, 1, outfile);

	for (j=0; j<d; j++) {
	  qtmp = ( Qnew_next[i] & (0x000000FF << 8*j) ) >> 8*j;
	  fwrite(&qtmp, sizeof qtmp, 1, outfile);
	}

	if (Qnew_data[i] == LAST) {
	  i++;
	  fwrite(&(Qnew_data[i]), sizeof qtmp, 1, outfile);
	  if (Qnew_data[i] == LAST)
	    fwrite(&(Qnew_next[i]), sizeof (unsigned int), 1, outfile);
	}
      }

      fwrite(&find_term_no, sizeof find_term_no, 1, outfile);
      for (i=0; i < find_term_no; i++) {
	fwrite_P(outfile, Mult_new[i][0]);
	}
      
      fclose(outfile);

      /* clear used data */
      Qnew_end = 0;
      Qnew_data[0] = LAST;
      Qnew_next[0] = MINUS_ONE;
      find_term_no = 0;
    }

    if (read_file) {
      if (depth != 0) {
	/* file open */
	strcpy(fileName, fileName_base);
	sprintf(depthName, "%d", depth);
	strcat(fileName, depthName);

	if (access(fileName, F_OK) != 0) {
	  /* If file not exists, then stop */
#ifdef DEBUG
	  fprintf(stderr, "File does not exists ! So we will stop. \n");
#endif
	  break;
	} else if ((outfile = fopen(fileName, "r")) == NULL) {
	  fprintf(stderr, "File open error !\n");
	  exit(2);
	}
	
	fread(&Qold_end, sizeof Qold_end, 1, outfile);
	
	for (i=0; i < Qold_end; i++) {
	  fread(&qtmp, sizeof qtmp, 1, outfile);
	  d = qtmp >> 12;
      
#ifdef DEBUG2
	  fprintf(stderr, "read %x", qtmp);
	  fprintf(stderr, " (qtmp >> 12 = %x, ", d);
#endif

	  Qold_data[i] = qtmp & LAST;
	  if (d == 0) Qold_next[i] = MINUS_ONE;
	  else Qold_next[i] = 0;
	  for (j=0; j < d; j++) {
	    fread(&qtmp, sizeof qtmp, 1, outfile);
	    Qold_next[i] += (qtmp << 8*j);
	  }

#ifdef DEBUG2
	  display_1qwt_tostderr(Qold_data[i]);
	  fprintf(stderr, "), Qold_next[i] = %x\n", Qold_next[i]);
#endif

	  if (Qold_data[i] == LAST) {
	    i++;
	    fread(&(Qold_data[i]), sizeof qtmp, 1, outfile);
	    if (Qold_data[i] == LAST)
	      fread(&(Qold_next[i]), sizeof (unsigned int), 1, outfile);

#ifdef DEBUG2
	    fprintf(stderr, " %x\n", Qold_data[i]);
	    fprintf(stderr, " %x\n", Qold_next[i]);
#endif

	  }
	}

	fread(&find_term_no, sizeof find_term_no, 1, outfile);
	for (i=0; i < find_term_no; i++) {
	  fread_P(outfile, Mult_old[i][0]);
	}
      
	Qnew_end = Qold_end;
	fclose(outfile);

#ifdef DEBUG
	time(&jikan);
	fprintf(stderr, "We finish the file reading at %s", ctime(&jikan));
#endif

      }
      
      /* display data when read_file = TRUE */
      get_qwt(0,0,just_display);

    } else {
      
    /* Step 1. Expand i-dominant monomials */

      get_qwt(0,0,develop_or_move);

    }

/*     debug_display(Qnew, Qnew_end - Qnew); */

    len_max = max(Qnew_end, len_max);
#ifdef DEBUG
    max_non_one = max(max_non_one, non_one);
    non_one = 0;
    /* So far, this is NOT what I want. I want to know
       non one l-weights, in Qnew. */
#endif

    if (!read_file && Qnew_end == 0) break;
    /* If find no new data, then end. */

#ifdef DEBUG
    time(&jikan);
    fprintf(stderr, "\n***** depth %d *****************\n", depth);
    fprintf(stderr, "%s", ctime(&jikan));
    memory_usage();
#endif

    /* Step 2. Write Qnew to the temporary file */

    /*    if (write_file2 && (dep_s == 0 || dep_s == depth+1)) { */
    if (write_file2) {

      /* file open */
      strcpy(fileName, fileName_base);
      sprintf(depthName, "%d", depth+1);
      strcat(fileName, depthName);

      strcat(fileName, "_tmp");
      if ((outfile2 = fopen(fileName, "w")) == NULL) {
	fprintf(stderr, "File open error !\n");
	exit(2);
      }

      fwrite(&Qnew_end, sizeof Qnew_end, 1, outfile2);

      for (i=0; i < Qnew_end; i++) {
	if (Qnew_next[i] == MINUS_ONE) 
	  d = 0;
	else if ((Qnew_next[i] & 0xFF000000) != 0)
	  d = 4;
	else if ((Qnew_next[i] & 0x00FF0000) != 0)
	  d = 3;
	else if ((Qnew_next[i] & 0x0000FF00) != 0)
	  d = 2;
	else
	  d = 1;

	qtmp = ( (Qnew_data[i] & LAST) | (d << 12) );
	fwrite(&qtmp, sizeof qtmp, 1, outfile2);

#ifdef DEBUG2
	fprintf(stderr, "write %x", qtmp);
	fprintf(stderr, "  (d = %x, ", d << 12);
	fprintf(stderr, "Qnew_data[i] = %x)\n", Qnew_data[i]);
#endif

	for (j=0; j<d; j++) {
	  qtmp = ( Qnew_next[i] & (0x000000FF << 8*j) ) >> 8*j;
	  fwrite(&qtmp, sizeof qtmp, 1, outfile2);
	}

	if (Qnew_data[i] == LAST) {
	  i++;
	  fwrite(&(Qnew_data[i]), sizeof qtmp, 1, outfile2);
	  if (Qnew_data[i] == LAST)
	    fwrite(&(Qnew_next[i]), sizeof (unsigned int), 1, outfile2);
	}
      }

      fwrite(&find_term_no, sizeof find_term_no, 1, outfile2);
      for (i=0; i < find_term_no; i++) {
	for (vertex=0; vertex<RANK; vertex++) {
	  fwrite_P(outfile2, Mult_new[i][vertex]);
	}
      }
      
      fclose(outfile2);

    }

    /* Step 3. Move found data */

    if (!read_file) { 
      for (i=0; i < Qnew_end; i++) {
	Qold_data[i] = Qnew_data[i];
	Qold_next[i] = Qnew_next[i];
      }
/*    memcpy(Qold, Qnew, (Qnew_end - Qnew)*sizeof (struct Qweights)); */

      for (i = 0; i < find_term_no; i++)
	subst_IP(Mult_old[i], Mult_new[i]);
/*    memcpy(Mult_old, Mult_new, find_term_no * RANK * sizeof (Poincare)); */

      Qold_end = Qnew_end;
      Qnew_end = 0;
      Qnew_data[0] = LAST;
      Qnew_next[0] = MINUS_ONE;
    }

#ifdef DEBUG
    max_num_terms = max(max_num_terms, find_term_no);
    MAX_exp = max(MAX_exp, max_exp);
    MAX_dim = max(MAX_dim, max_dim);
#ifdef EULER
    MAX_euler = max(MAX_euler, max_euler);
#endif
    max_exp = 0; max_dim = 0;
#ifdef EULER
    max_euler = 0;
#endif

#endif
    find_term_no = 0;

  } while (++depth < DEPTH_MAX);

#ifdef DEBUG
  time(&jikan);
  fprintf(stderr, "\n***** TOTAL = depth %d ************\n", depth);
  fprintf(stderr, "%s", ctime(&jikan));
  fprintf(stderr, "\n***** MAXIMUM OF ");
  Qnew_end = len_max; find_term_no = max_num_terms;  max_exp = MAX_exp;
#ifdef EULER
  max_euler = MAX_euler;
#endif
  if (DIM_range.low != -1 || DIM_range.high != -1)
      MAX_dim = max_dim;
  memory_usage();
#endif
  exit(0);

}
