#include "stdlib.h"
#include "ctype.h"
#include "sys/types.h"
#include "qchar.h"
#include "time.h"
#include <unistd.h>

#define display_P(p) display_P_tofile(stdout, p)
#define display_PL(p) display_PL_tofile(stdout, p)

struct Range DIM_range;
int opt_max_exp;
boolean IS_L_DOM, IS_DOM, EQ_WEIGHT;
boolean print_depth = FALSE;

Weight WEIGHT;
Wttable highestweight;
Qweight hwt;

Qweight qwt_develop;
Qweight_cont Qold_data[LEN_MAX];
unsigned int Qold_end;
unsigned int Qold_next[LEN_MAX];
Poincare Mult_old[MAX_NUM_TERMS][RANK];

Weight FoundWei[CHAR_MAX];
PoincareLong FoundPoi[CHAR_MAX];

char *prog;
Poincare pone, pzero; /* constant Poincare polynomials */

unsigned short int depth = 0;

FILE *infile;

unsigned int find_weight_no = 0;
unsigned int find_term_no = 0;

extern void read_hwt(char *str);
extern void str2weight(char *str);
extern void dim_range(char *str, struct Range *range);

extern void display_table_tofile(FILE *file, Wttable table);
#define display_table(table) display_table_tofile(stdout, table)
#define display_table_tostderr(table) display_table_tofile(stderr, table)
extern void display_table_tostr(Wttable table, char *string);

extern void display_weight(Weight weight);
extern boolean cmp_weight(Weight weight1, Weight weight2);

extern void get_qwt(unsigned short int ii, unsigned int address, void (*p_func)(unsigned int));

extern Vertex iota(Vertex no);

void usage()
{
  fprintf(stderr,"Usage:  %s [-options] l-highest_weight\n"
	  "  l-highest weight is a sequence of a triple = vertex,eigen,expo\n"
	  "  e.g. 1,0,1,1,1,2 = Y_{1,a} Y_{1,qa}^2 \n"
	  "Options:\n"
	  "   -p      : print dominants weights only\n"
	  "   -w#,#,#,...,# (rank times)\n"
	  "           : print given weights "
	  "(# can be a negative integer)\n"
	  "   -h      : this help\n"
	  "REMARK: become super user and unlimit stacksize, for some cases\n"
	  "Vertex Numbering:\n",
	  prog);
  show_vertex();
  exit(1);
}

#ifdef DEBUG
void develop_error(Vertex vertex, Wttable table, Poincare coeff,\
		   Poincare p[RANK])
{
  Vertex v;

  fprintf(stderr, "Error in algorithm at vertex %d. Coeff = ",
	  vertex+1);
  display_P_tofile(stderr, coeff);
  fprintf(stderr, "\n  Qweight = "); display_table_tostderr(table);
  fprintf(stderr, "\n  Virtual Poincare polynomials are:\n  ");
  for (v=0; v < RANK; v++) {
    display_P_tofile(stderr, p[v]);
    fprintf(stderr, "   ");
  }
  fprintf(stderr, "\n");
  exit(3);
}
#endif

void find_weight(unsigned int address)
{
  Poincare coeff,p[RANK];
  Weight weight;

  PoincareLong coeffL;
  unsigned short int t;

  Wttable vtable,table;
  /* boolean already_i_dom = FALSE; */

  Vertex vertex;
  unsigned int _term_no;
  unsigned int _weight_no;

  qwt2vtable(qwt_develop,vtable);
  U(vtable,table);
  table2weight(table,weight);

  if (IS_DOM && !is_dominant(table)) return;

  if (depth == 0) subst_P(coeff,pone);

  if (Qold_data[address] == LAST) {
    _term_no = Qold_next[address];
    subst_P(coeff, Mult_old[_term_no][0]);
  } else {

    /* In this case, coeff must be pone */
    subst_P(coeff,pone);

    /* But we check it */
#ifdef DEBUG
    byte2poincare(Qold_data[address],p);
    for (vertex=0; vertex < RANK; vertex++) {
      if (!is_i_dom(table,vertex)) {
	if (!equal_P(coeff,p[vertex]))
	  develop_error(vertex,table,coeff,p);
      }
    }
#endif	
  }

  _weight_no = 0;

  /* coeff. no kata henkou */
#ifdef EULER
  coeffL = (unsigned long int)coeff;
#else
  for (t=0; t<COMP_DIM_MAX; t++)
    coeffL[t] = (unsigned long int)coeff[t];
#endif

  while (_weight_no < find_weight_no) {

    if (cmp_weight(FoundWei[_weight_no],weight)) {

      /* if weight is found already */
      subadd_PL(FoundPoi[_weight_no],coeffL);

#ifdef DEBUG2
      printf("Found Before # %d :", _weight_no);
      printf("Weight = ");
      display_weight(weight);
      printf(", multi. = ");
      display_P(coeff);
      printf(", ");
      printf("\n");
#endif

      return;

    }
    _weight_no++;
  } 

  /* weight is not found before */

  for (vertex=0; vertex < RANK; vertex++)
    FoundWei[find_weight_no][vertex] = weight[vertex];
  subst_PL(FoundPoi[find_weight_no],coeffL);
  
#ifdef DEBUG2
  printf("# %d :",find_weight_no);
  printf("Weight = ");
  display_weight(weight);
  printf(", multi. = ");
  display_P(coeff);
  printf(", ");
  printf("\n");
#endif

  find_weight_no++;

  if (find_weight_no >= CHAR_MAX) {
    fprintf(stderr,\
	    "Out of Memory at depth = %d. Redefine CHAR_MAX larger.\n",\
	    depth);
    exit(2);
  }
}

main(int argc, char *argv[])
{

  char fileName[40] = "          ", fileName_base[40] = "    ";
  /*  char *fileName, *fileName_base; */
  char depthName[3];
  Vertex vertex;

  Weight tmp_weight, tmp_weight2;

  unsigned int i,j,d;
  unsigned short int dep_s;

  Qweight_cont qtmp;

#ifdef DEBUG
  time_t jikan;
#endif

  /*  
  fileName = (char *)malloc(40);
  fileName_base = (char *)malloc(40);
  */

  /* Set pone = 1, pzero = 0 */
  setp10();

  DIM_range.low = -1;
  DIM_range.high = -1;
  opt_max_exp = -1;
  IS_L_DOM = FALSE;
  EQ_WEIGHT = FALSE;

  /* Option handling */

  prog = *argv++;
  while (--argc > 1) {
    if (argv[0][0] == '-') {
      switch (argv[0][1]) {
      case 'D':
	dim_range(*argv+2,&DIM_range);
	/* printf("%d - %d\n", DIM_range.low,DIM_range.high); */
	break;
      case 'd':
	print_depth = TRUE;
	break;
      case 'p':
	IS_DOM = TRUE;
	break;
      case 'w':
	EQ_WEIGHT = TRUE;
	str2weight(*argv+2);
	/* for (i=0; i<RANK; i++)
	   printf("%d ",WEIGHT[i]);
	   printf("\n"); */
	break;
      case 'h':
      default:
	usage();
      }
      argv++;
    }
    else usage();
  }

  /* read highest weight */
  if (*argv == NULL) usage();
  read_hwt(*argv);

  /* display highest weight */
  /*    to stderr */
  fprintf(stderr,"*** highest weight =");
  display_table_tostderr(highestweight);
  fprintf(stderr," ***\n");

  Qold_data[0] = LAST;
  Qold_next[0] = MINUS_ONE;

  Qold_end = 0;

  /* file name */
  strncpy(fileName_base, prog, 4); /* take "./E8" and cut _q2c */
  /* strcpy(fileName_base, "./E8"); */
  display_table_tostr(highestweight,depthName);
  strcat(fileName_base, depthName);
  strcat(fileName_base, " ");

  if (EQ_WEIGHT) {
    qwt2weight(hwt,tmp_weight);
    for (vertex=0; vertex < RANK; vertex++) {
      tmp_weight[vertex] -= WEIGHT[vertex];
#ifdef DEBUG2
      fprintf(stderr,"Vertex = %d, Weight = %d\n", vertex, tmp_weight[vertex]);
#endif
    }
    if (! Cinv(tmp_weight,tmp_weight2)) {
      fprintf(stderr,"No vectors of that weight !\n");
#ifdef DEBUG2
      for (vertex=0; vertex < RANK; vertex++) 
	fprintf(stderr,"Vertex = %d, weight = %d\n", vertex, tmp_weight2[vertex]);
#endif
      exit(1);
    }

    for (vertex=0; vertex < RANK; vertex++)
      depth += tmp_weight2[vertex];

#ifdef DEBUG2
    fprintf(stderr, "depth = %d\n", depth);
#endif    

  }

  /* Calculate dep_s = the depth when we end */
  qwt2weight(hwt,tmp_weight);
  for (vertex=0; vertex < RANK; vertex++) 
    tmp_weight2[vertex] = tmp_weight[iota(vertex)];
  for (vertex=0; vertex < RANK; vertex++) 
    tmp_weight[vertex] += tmp_weight2[vertex];
  Cinv(tmp_weight,tmp_weight2);
  dep_s = 0;
  for (vertex=0; vertex < RANK; vertex++) {
#ifdef DEBUG2
    fprintf(stderr,"%d ",tmp_weight2[vertex]);
#endif
    dep_s += tmp_weight2[vertex];
  }

#ifdef DEBUG2
  fprintf(stderr, "dep_s = %d\n", dep_s);
#endif


  /* Main Routine */  

  do {

    if (depth != 0) {
      /* file open */
      strcpy(fileName, fileName_base);
      sprintf(depthName, "%d", depth);
      strcat(fileName, depthName);

      if (access(fileName, F_OK) != 0) {
	/* If file not exists, then stop */
	if (depth != dep_s+1) {
	  fprintf(stderr, "File %s does not exists ! So we stop. \n", fileName);
	  strncpy(fileName_base, prog, 4); /* take "./E8" and cut _q2c */
	  fileName_base[4] = '\0';
	  fprintf(stderr, "Run %s with the option -f. \n",fileName_base);
	  exit(2);
	}
	break;

      } else if ((infile = fopen(fileName, "r")) == NULL) {
	fprintf(stderr, "File open error !\n");
	exit(2);
      }
	
      fread(&Qold_end, sizeof Qold_end, 1, infile);

      for (i=0; i < Qold_end; i++) {
	fread(&qtmp, sizeof qtmp, 1, infile);
	d = qtmp >> 12;
      
#ifdef DEBUG2
	fprintf(stderr, "read %x", qtmp);
	fprintf(stderr, " (qtmp >> 12 = %x, ", d);
#endif

	Qold_data[i] = qtmp & LAST;
	if (d == 0) Qold_next[i] = MINUS_ONE;
	else Qold_next[i] = 0;
	for (j=0; j < d; j++) {
	  fread(&qtmp, sizeof qtmp, 1, infile);
	  Qold_next[i] += (qtmp << 8*j);
	}

#ifdef DEBUG2
	display_1qwt_tostderr(Qold_data[i]);
	fprintf(stderr, "), Qold_next[i] = %x\n", Qold_next[i]);
#endif

	if (Qold_data[i] == LAST) {
	  i++;
	  fread(&(Qold_data[i]), sizeof qtmp, 1, infile);
	  if (Qold_data[i] == LAST)
	    fread(&(Qold_next[i]), sizeof (unsigned int), 1, infile);

#ifdef DEBUG2
	  fprintf(stderr, " %x\n", Qold_data[i]);
	  fprintf(stderr, " %x\n", Qold_next[i]);
#endif

	}
      }
    
      fread(&find_term_no, sizeof find_term_no, 1, infile);
      for (i=0; i < find_term_no; i++) {
	fread_P(infile, Mult_old[i][0]);
      }
    
      fclose(infile);

#ifdef DEBUG
      time(&jikan);
      fprintf(stderr, "We finish the file reading at %s", ctime(&jikan));
#endif

    }
  
    if (print_depth) 
      printf("\n***** depth %d *****************\n", depth);
#ifdef DEBUG      
    else
      fprintf(stderr,"\n***** depth %d *****************\n", depth);
#endif

    find_weight_no = 0;

    get_qwt(0,0,find_weight);

#ifdef DEBUG
    fprintf(stderr, "We found #%d", find_weight_no);
    if (IS_DOM) fprintf(stderr, " dominant");
    fprintf(stderr, " weights.\n");
#endif

    for (i=0; i < find_weight_no; i++) {
    
      if (!EQ_WEIGHT || cmp_weight(FoundWei[i],WEIGHT)) {
	printf("Weight = ");
	display_weight(FoundWei[i]);
	printf(", multi. = ");
	display_PL(FoundPoi[i]);
	printf(", ");
	printf("\n");
      }

    }
    
    /* if `-w' is set, we do not go loop. */
    if (EQ_WEIGHT) break;

    /* if `-p' is set and depth > dep_s/2, then we stop. */
    if (IS_DOM && depth*2 >= dep_s) break;

  } while (++depth < DEPTH_MAX);

}
