#include "stdlib.h"
#include "ctype.h"
#include "sys/types.h"
#include "qchar.h"

extern Poincare Mult[MAX_NUM_TERMS][RANK];
extern Poincare Mult2[MAX_NUM_TERMS][RANK];
extern void memory_usage();
extern unsigned short int depth;
extern Wttable highestweight;
extern Qweight hwt;

extern Qweight qwt_develop;
extern Qweight_cont Qold_data[LEN_MAX];
extern unsigned int Qold_end;
extern unsigned int Qold_next[LEN_MAX];


extern void mul_A_sub(Wttable table,Vertex vertex,Eigen eigen,short int add);
extern short int exponent(Wttable vtable, Vertex vertex, Eigen eigen);

void get_qwt(unsigned short int ii, unsigned int address, void (*p_func)(unsigned int))
     /* p_func : function pointer 'develop_or_move' or 'move_dep_zero' or
      'just_display' */
{

  do {
    
    if ((qwt_develop[ii] = Qold_data[address]) == LAST) {
      /* get qwt ! */ 
      /* develop_or_move(address+1); */
      (*p_func)(address+1);
      /* address is necessary only to get term_no. */
    } else {
      get_qwt(ii+1, address+1, p_func);
    }

    if (Qold_next[address] == MINUS_ONE) /* when there are no more data */
      break;

    address += Qold_next[address];
  } while(TRUE);
}

void U(Wttable vtable, Wttable table)
     /* calculating u = (q-analog of) w - Cv */
{
  Vertex vertex;
  Eigen eigen;

  for (vertex = 0; vertex < RANK; vertex++) 
    for (eigen = 0; eigen < EIG_MAX; eigen++)
      table[vertex][eigen] = exponent(vtable, vertex, eigen);
}

#ifdef COMMENT
/* The next three subroutines return vtable = C^{-1}(w-u) (q-analog)
   for given u = table, w = highestweight. */
void u2v_sub(Eigen eigen, Wttable table)
{
  Vertex vertex;
  Wttable tmp;

  sub_table(tmp,table);
  for (vertex=0; vertex<RANK; vertex++) {
    mul_A_sub(tmp,vertex,eigen,table[vertex][eigen]);
    if (eigen > 0)
      tmp[vertex][eigen+1] -= table[vertex][eigen-1];
  }
  sub_table(table,tmp);
}

void u2v(Wttable table)
{
  Wttable tmp;

  Vertex i;
  Eigen j;

  clear_table(tmp);
  for (i=0; i < RANK; i++)
    tmp[i][1] += table[i][0];
  
  for (j=1; j < EIG_MAX-1; j++) {
    u2v_sub(j-1,tmp);
    for (i=0; i < RANK; i++)
      tmp[i][j+1] += table[i][j];
  }
  sub_table(table,tmp);
}

void V(Wttable table, Wttable vtable)
{
  Vertex vertex;
  Eigen j;

  sub_table(vtable,highestweight);
  for (vertex = 0; vertex < RANK; vertex++)
    for (j = 0; j < EIG_MAX; j++)
      vtable[vertex][j] -= table[vertex][j];
  u2v(vtable);
}
#endif

/* calculate dimension of graded quiver variety
   Remark: dim of $\M(\rho)$, not $\M(\rho)\cap\La(\bw)$. */
unsigned int dim(Wttable table, Wttable vtable)
{
 
  Vertex i;
  Eigen j;
  unsigned int d = 0;
  
  /* printf("highest_weight : "); display_table(highestweight); printf("\n");
  printf("vtable         : "); display_table(vtable); printf("\n");
  printf("C_{k,lambda}   : "); display_table(table); printf("\n"); */
  
  for (i = 0; i < RANK; i++)
    for (j = 1; j < EIG_MAX; j++)
      d += (table[i][j+1] + highestweight[i][j-1])* vtable[i][j];
  return(d);
}

#ifdef COMMENT
void table2qwt(Wttable table, Qweight qwt)
{
  Vertex i;
  Eigen j;
  unsigned short int c = 0;

  Wttable vtable;

  V(table, vtable);

  for (j = 0; j < EIG_MAX; j++)
    for (i = 0; i < RANK; i++)
      if (vtable[i][j] != 0) {
	qwt[c++] = iexpo(vtable[i][j]) + ieigen(j) + i;
 	if (c + 1 > DEPTH_MAX) {
	  fprintf(stderr,\
		  "Out of range in a q-weight table at depth = %d. "
		  "Redefine DEPTH_MAX larger\n", depth);
	  memory_usage();
	  exit(1);
	}
      }
  qwt[c] = 0;
}
#endif

#ifdef COMMENT
void vtable2qwt(Wttable vtable, Qweight qwt)
{
  Vertex i;
  Eigen j;
  unsigned short int c = 0;
  unsigned short int a;

  for (j = 0; j < EIG_MAX; j++)
    for (i = 0; i < RANK; i++)
      for (a = vtable[i][j]; a > 0; a--) {
	qwt[c++] = ieigen(j) + i;
	if (c + 1 > DEPTH_MAX) {
	  fprintf(stderr,\
		  "Out of range in a q-weight table at depth = %d. "
		  "Redefine DEPTH_MAX larger\n", depth);
	  memory_usage();
	  exit(1);
	}
      }
  qwt[c] = LAST;
}
#endif

#ifdef COMMENT
void qwt2table(Qweight qwt, Wttable table)
{
  int i;
  Wttable vtable;

  clear_table(vtable);
  i = 0;
  while (qwt[i] != 0) {
    vtable[vertexf(qwt[i])][eigenf(qwt[i])] = expo(qwt[i]);
    i++;
  }

  U(vtable, table);
/*   sub_table(table, vtable); */
}
#endif

void qwt2vtable(Qweight qwt, Wttable vtable)
{
  unsigned short int i;

  clear_table(vtable);
  i = 0;
  while (qwt[i] != LAST) {
    vtable[vertexf(qwt[i])][eigenf(qwt[i])]++;
    i++;
  }
}

/* display table */
void display_table_tofile(FILE *file, Wttable table)
{
  Vertex i;
  Eigen j;
  int a;
  boolean is_one = TRUE;

  for (i=0; i<RANK; i++)
    for (j=0; j < EIG_MAX; j++)
      if ((a = table[no_vertex(i+1)][j]) != 0) {
	is_one = FALSE;
	if (j > 9) fprintf(file, " %d_{%d}",i+1,j);
	else fprintf(file, " %d_%d",i+1,j);
	if (a > 1 && a < 10) fprintf(file, "^%d", a);
	else if (a != 1) fprintf(file, "^{%d}",a);
      }
  if (is_one) fprintf(file, " *");
}

#define display_table(table) display_table_tofile(stdout, table)

void display_qwt(Qweight qwt)
{
  Wttable table;

  qwt2vtable(qwt, table);
  display_table(table);

}

#if COMMENT
void display_qwt2(Qweight qwt, unsigned int _term_no)
{
  Wttable table;
  Vertex vertex;

  qwt2table(qwt, table);
  display_table(table);
  printf("\n");

}
#endif

void display_1qwt_tostderr(Qweight_cont data)
{
  Vertex i;
  Eigen j;
/*   int a; */

  if ((data & LAST) == LAST) fprintf(stderr, "*");
  else {
    i = vertexf(data);
    j = eigenf(data);
/*     a = expo(data); */

    if (j > 9) fprintf(stderr, "%d_{%d}",i+1,j);
    else fprintf(stderr, "%d_%d",i+1,j);
/*     if (a > 1 && a < 10) fprintf(stderr, "^%d", a); */
/*     else if (a != 1) fprintf(stderr, "^{%d}",a); */
  }
}

void display_qwt_tostderr(Qweight qwt)
{
  unsigned short i = 0;

  while (qwt[i] != LAST) {
    display_1qwt_tostderr(qwt[i++]);
    fprintf(stderr, " ");
  }
}

/* display table to string */
void display_table_tostr(Wttable table, char *string)
{
  Vertex i;
  Eigen j;
  short int a;
  boolean is_one = TRUE;
  char temp1[20] = "", temp2[20];

  for (i=0; i<RANK; i++)
    for (j=0; j < EIG_MAX; j++)
      if ((a = table[no_vertex(i+1)][j]) != 0) {
	is_one = FALSE;
	if (j > 9) sprintf(temp2," %d_{%d}",i+1,j);
	else sprintf(temp2," %d_%d",i+1,j);
	strcat(temp1,temp2);
	if (a > 1 && a < 10) {
	  sprintf(temp2,"^%d", a); strcat(temp1,temp2);
	}
	else if (a != 1) {
	  sprintf(temp2,"^{%d}",a);  strcat(temp1,temp2);
	}
      }
  if (is_one) sprintf(temp1," *");
  strcpy(string, temp1);
}

/* Subroutines treating data type Weight, Qweight, .. */

void sub_qwt(Qweight qwt1, Qweight qwt2)
{
  unsigned short int i = 0;
  Qweight_cont a;

  while ((a = qwt2[i]) != LAST) {
    qwt1[i++] = a;
  }
  qwt1[i] = LAST;
}

unsigned short int qwt_len(Qweight qwt)
{
  unsigned short int i = 0;

  while (qwt[i++] != LAST)
    ;
  
  return(i);
}

unsigned short int qwt_insert(Qweight qwt, Eigen eigen, Vertex vertex)
{
  unsigned short int i = 0;
  Qweight_cont a;

  while ((a = qwt[i]) != LAST && eigenf(a) < eigen)
    i++;

  if (a != LAST) {
    while ((a = qwt[i]) != LAST && eigenf(a) == eigen\
	   && vertexf(a) < vertex)
      i++;
  }

  return(i);

  /*
#ifdef DEBUG
  if (i == DEPTH_MAX - 1) {
    fprintf(stderr,\
	    "Out of range in a q-weight table at depth = %d. "
	    "Redefine DEPTH_MAX larger\n", depth);
    memory_usage();
    exit(1);
  }
#endif

  memmove(qwt+i+1, qwt+i, (DEPTH_MAX-i-1)*sizeof (Qweight_cont));
  qwt[i] = ieigen(eigen) + vertex;

  */
}    

#ifdef COMMENT
clear_qwt(qwt)
     Qweight qwt;
{
  unsigned short int i;

  for (i=0; i < DEPTH_MAX; i++) {
    qwt[i].vertex = 0;
    qwt[i].eigen = 0;
    qwt[i].expo = 0;
  }
}

/*
void sub_qwt(Qweight_cont *qwt1, Qweight_cont *qwt2)
{

  while ((*qwt1++ = *qwt2++) != 0);

}
*/

boolean equal_qwt(Qweight qwt1, Qweight qwt2)
     /* struct Qweight_cont *qwt1, *qwt2; */
{

  unsigned int i = 0;
  Qweight_cont a;

  while ((a = qwt1[i]) != 0)
    if (a != qwt2[i++])
      return(FALSE);
  if (qwt2[i] == 0) return(TRUE);
  else return(FALSE);

  /*
  if (strcmp((char *)qwt1, (char *)qwt2)!=0) return(FALSE);
  else return(TRUE);
  */
}

boolean equal_qwt(Qweight_cont *qwt1, Qweight_cont *qwt2)
     /* struct Qweight_cont *qwt1, *qwt2; */
{

  Qweight_cont a;

  while ((a = *qwt1++) != 0)
    if (a != *qwt2++)
      return(FALSE);
  if (*qwt2 == 0) return(TRUE);
  else return(FALSE);

  /*
  if (strcmp((char *)qwt1, (char *)qwt2)!=0) return(FALSE);
  else return(TRUE);
  */

}

/* check from qwt1[2] */
boolean equal_qwt2(Qweight qwt1, Qweight qwt2)
{

  unsigned int i = 2;
  Qweight_cont a;

  while ((a = qwt1[i]) != 0)
    if (a != qwt2[i++])
      return(FALSE);
  if (qwt2[i] == 0) return(TRUE);
  else return(FALSE);
}
#endif

unsigned int exp_max(Wttable table)
     /* return 0 when all exp =< 0 */
{
  Vertex i;
  Eigen j;
  int m = 0;

  for (i=0; i<RANK; i++)
    for (j=0; j<EIG_MAX; j++)
      m = max(table[i][j],m);
  return(m);
}

boolean equal_weight(Wttable table, Weight weight)
{
  Vertex vertex;
  Eigen j;
  int s;

  for (vertex = 0; vertex < RANK; vertex++) {
    s = 0;
    for (j = 0; j < EIG_MAX; j++)
      s += table[vertex][j];
    if (s != weight[vertex]) return(FALSE);
  }
  return(TRUE);
}

boolean cmp_weight(Weight weight1, Weight weight2)
{
  Vertex vertex;
 
  for (vertex = 0; vertex < RANK; vertex++) {
    if (weight1[vertex] != weight2[vertex]) return(FALSE);
  }
  return(TRUE);
}

void table2weight(Wttable table, Weight weight)
{
  Vertex vertex;
  Eigen eigen;

  for (vertex=0; vertex < RANK; vertex++) {
    weight[vertex] = 0;
    for (eigen = 0; eigen < EIG_MAX; eigen++)
      weight[vertex] += table[vertex][eigen];
  }
}

void qwt2weight(Qweight qwt, Weight weight)
{
  Vertex vertex;
  Eigen eigen;
  Wttable vtable;
  
  qwt2vtable(qwt,vtable);
  table2weight(vtable,weight);

}

boolean is_i_dom(Wttable table, Vertex vertex)
{
  Eigen eigen;

  for (eigen=0; eigen < EIG_MAX; eigen++) {
    if (table[vertex][eigen] < 0) return(FALSE);
  }
  return(TRUE);
}

#ifdef COMMENT
boolean is_i_dom2(Wttable vtable, Vertex vertex)
{
  Eigen eigen;

  for (eigen=0; eigen < EIG_MAX; eigen++) {
    if (exponent(vtable, vertex, eigen) < 0) return(FALSE);
  }
  return(TRUE);
}
#endif

/*
boolean is_non_i_dom(Wttable table,Vertex vertex)
{
  Eigen eigen;

  for (eigen=0; eigen < EIG_MAX; eigen++) {
    if (table[vertex][eigen] < 0) return(TRUE);
  }
  return(FALSE);
}
*/

boolean is_l_dominant(Wttable table)
{
  Vertex vertex;

  for (vertex = 0; vertex < RANK; vertex++)
    if (!is_i_dom(table, vertex)) return(FALSE);
  return(TRUE);
}

boolean is_i_regdom(Wttable table,Vertex vertex)
{
  Eigen eigen;
  boolean detekita = FALSE;

  for (eigen=0; eigen<EIG_MAX; eigen++) {
    if (table[vertex][eigen] < 0) return(FALSE);
    if (table[vertex][eigen] > 0) detekita = TRUE;
  }
  if (detekita) return(TRUE);
  return(FALSE);
}

int weight_i(Vertex vertex, Wttable table)
{
  int s,j;

  s = 0;
  for (j = 0; j < EIG_MAX; j++) s += table[vertex][j];
  return(s);
}


boolean is_dominant(Wttable table)
{
  Vertex vertex;

  for (vertex = 0; vertex < RANK; vertex++)
    if (weight_i(vertex, table) < 0) return(FALSE);
  return(TRUE);
}

#ifdef COMMENT
boolean is_anti_dominant(Wttable table)
{
  Vertex vertex;

  for (vertex = 0; vertex < RANK; vertex++)
    if (weight_i(vertex, table) > 0) return(FALSE);
  return(TRUE);
}
#endif

void display_weight(Weight weight)
{
  Vertex i;
  int a;
  boolean is_one = TRUE;

  for (i=0; i<RANK; i++)
    if ((a = weight[no_vertex(i+1)]) != 0) {
      is_one = FALSE;
      printf(" %d", i+1);
      if (a > 1 && a < 10) printf("^%d", a);
      else if (a != 1) printf("^{%d}",a);
    }
  if (is_one) printf(" e^0");
}

#ifdef COMMENT
int display_add_qwt(Qweight qwt)
{
  int i = 0;
  int exp;

  while (qwt[i] != 0) {
    if (eigenf(qwt[i]) > 9)
      printf(" %d_{%d}", vertexf(qwt[i])+1, eigenf(qwt[i]));
    else
      printf(" %d_%d", vertexf(qwt[i])+1, eigenf(qwt[i]));
    if ((exp = expo(qwt[i])) > 1 && exp < 10) {
      printf("^%d", exp);
    } else if (exp != 1)
      printf("^{%d}", exp);
    i++;
  }
  if (i == 0) printf(" *");
  /* printf("\n"); */
  return(i+1);
}

int add_qwt(Qweight qwt)
{
  int i = 0;

  while (qwt[i++] != 0);
  return(i);
}

int subadd_qwt(Qweight_cont *qwt1, Qweight_cont *qwt2)
{
  Qweight_cont *qwt3;

  qwt3 = qwt1;
  while (*qwt2 != 0) {
    *qwt1++ = *qwt2++;
  }
  *qwt1++ = 0;
  return(qwt1-qwt3);
}
#endif

